@@ -162,17 +162,103 @@ def e57_fix_rotation_order(rotation_matrix:np.array) -> np.array:
             [r[1,0],r[1,1],-r[1,2]],
             [r[2,0],r[2,1],-r[2,2]]]
 
+def generate_virtual_images(geometries: List[o3d.geometry.Geometry],cartesianTransforms: List[np.array],width:int=640,height:int=480,f:float=400)-> o3d.geometry.Image:
+    """Generate a set of Open3D Images from cartesianTransforms and geometries. 
+    The same intrinsic camera parameters are used for all cartesianTransforms that are passed to the function
+
+    .. image:: ../../pics/rendering3.PNG
+
+    Args:
+        geometries (List[o3d.geometry]):o3d.geometry.PointCloud or o3d.geometry.TriangleMesh
+        cartesianTransforms (List[np.array 4x4]): [Rt]
+        width (int, optional): image width in pix. Defaults to 640.
+        height (int, optional): image height in pix. Defaults to 480.
+        f (float, optional): focal length in pix. Defaults to 400.
+
+    Returns:
+        List[o3d.cpu.pybind.geometry.Image]
+    """
+    #set renderer
+    render = o3d.visualization.rendering.OffscreenRenderer(width,height)
+    mtl=o3d.visualization.rendering.MaterialRecord()
+    mtl.base_color = [1.0, 1.0, 1.0, 1.0]  # RGBA
+    mtl.shader = "defaultUnlit"
+
+    # set internal camera orientation
+    fx=f
+    fy=f
+    cx=width/2-0.5
+    cy=width/2-0.5
+    intrinsic=o3d.camera.PinholeCameraIntrinsic(width,height,fx,fy,cx,cy)
+    
+    #add geometries
+    geometries=ut.item_to_list(geometries)
+    for idx,geometry in enumerate(geometries):
+        render.scene.add_geometry(str(idx),geometry,mtl) 
+
+    #set cameras and generate images
+    list=[]
+    for cartesianTransform in cartesianTransforms:
+        extrinsic=np.linalg.inv(cartesianTransform)
+        render.setup_camera(intrinsic,extrinsic)
+        img = render.render_to_image()
+        list.append(img)
+
+    if len(list) !=0:
+        return list
+    else:
+        return None
+
+def generate_virtual_image(geometries: List[o3d.geometry.Geometry],pinholeCamera: o3d.camera.PinholeCameraParameters)-> o3d.geometry.Image:
+    """Generate an Open3D Image from a set of geometries and an Open3D camera (). 
+
+    Args:
+        geometries (List[o3d.geometry]):o3d.geometry.PointCloud or o3d.geometry.TriangleMesh
+        pinholeCamera (o3d.camera.PinholeCameraParameters): extrinsic (cartesianTransform) and intrinsic (width,height,f,principal point U and V) camera parameters
+
+    Returns:
+        o3d.cpu.pybind.geometry.Image or None
+    """
+    #create renderer
+    width=pinholeCamera.intrinsic.width
+    height=pinholeCamera.intrinsic.height
+    render = o3d.visualization.rendering.OffscreenRenderer(width,height)
+
+    # Define a simple unlit Material. (The base color does not replace the geometry colors)
+    mtl=o3d.visualization.rendering.MaterialRecord()
+    mtl.base_color = [1.0, 1.0, 1.0, 1.0]  # RGBA
+    mtl.shader = "defaultUnlit"
+
+    #set camera
+    render.setup_camera(pinholeCamera.intrinsic,pinholeCamera.extrinsic)
+
+    #add geometries
+    geometries=ut.item_to_list(geometries)
+
+    for idx,geometry in enumerate(geometries):
+        render.scene.add_geometry(str(idx),geometry,mtl) 
+
+    #render image
+    img = render.render_to_image()
+
+    if img is not None:
+        return img
+    else:
+        return None
+
+
 def crop_geometry(geometry:o3d.geometry, box:o3d.geometry.OrientedBoundingBox, subdivide:int = 0) ->o3d.geometry.TriangleMesh:
     """"Crop portion of a mesh/pointcloud that lies within an OrientedBoundingBox 
 
+    .. image:: ../../pics/selection_BB_mesh2.PNG
+
     Args:
         geometry (o3d.geometry.TriangleMesh or o3d.geometry.PointCloud)
         box (o3d.geometry.OrientedBoundingBox )
         subdivide (int): number of interations to increase the density of the mesh (1=x4, 2=x16, etc.)
         
     Returns:
-        mesh:       o3d.geometry.TriangleMesh 
-        None:       if no data was retained
+        o3d.geometry.TriangleMesh or None
     """
     # transform box to axis aligned box 
     r=box.R
@@ -206,8 +292,6 @@ def crop_geometry(geometry:o3d.geometry, box:o3d.geometry.OrientedBoundingBox, s
     else:
         return None
 
-
-
 def expand_box (box: o3d.geometry.OrientedBoundingBox, u=5.0,v=5.0,w=5.0) -> o3d.geometry.OrientedBoundingBox:
     """expand an o3d.geometry.OrientedBoundingBox in u(x), v(y) and w(z) direction with a certain offset
 
@@ -429,6 +513,16 @@ def ifc_to_mesh(ifcElement:ifcopenshell.entity_instance)-> o3d.geometry.Triangle
         return None
   
 
+
+def save_view_point(geometry, filename):
+    vis = o3d.visualization.Visualizer()
+    vis.create_window()
+    vis.add_geometry(geometry)
+    vis.run()  # user changes the view and press "q" to terminate
+    param = vis.get_view_control().convert_to_pinhole_camera_parameters()
+    o3d.io.write_pinhole_camera_parameters(filename, param)
+    vis.destroy_window()
+
 def get_center(mesh:o3d.geometry.TriangleMesh,triangle:np.array) -> np.array:
     """
     Compute center of an Open3D mesh face
@@ -545,6 +639,14 @@ def get_translation(cartesianTransform:np.array) -> np.array:
         return np.array([cartesianTransform[0,3],cartesianTransform[1,3],cartesianTransform[2,3]])
     else:
         return None
+    
+def get_rotation(cartesianTransform:np.array) -> np.array:
+    if cartesianTransform.size ==16:
+        return np.array([[cartesianTransform[0,1],cartesianTransform[0,2],cartesianTransform[0,3]],
+                        [cartesianTransform[1,1],cartesianTransform[1,2],cartesianTransform[1,3]],
+                        [cartesianTransform[2,1],cartesianTransform[2,2],cartesianTransform[2,3]]])
+    else:
+        return None
 
 def lambert72_to_spherical_coordinates(x :float ,y:float) -> tuple: 
     """"
