@@ -3,20 +3,18 @@ ImageNode - a Python Class to govern the data and metadata of image data (JPG,PN
 """
 #IMPORT PACKAGES
 import xml.etree.ElementTree as ET
-
 import cv2
 import PIL
-import PIL.Image as PILimage
-from PIL import ImageDraw, ImageEnhance, ImageFont
 from PIL.ExifTags import GPSTAGS, TAGS
 from rdflib import Graph, URIRef
-import uuid    
 import numpy as np
 import os
 #IMPORT MODULES
 from geomapi.node import Node
 import geomapi.utils as ut
 import geomapi.geometryutils as gt
+import open3d as o3d
+import math
 
 SUPPORTED_POINT_FIELDS = {
     "xResolution": "(float)",
@@ -105,9 +103,23 @@ class ImageNode(Node):
                 self.timestamp=ut.get_timestamp(graphPath)
                 self.get_metadata_from_graph_path() 
             else:
-                raise ValueError('file must be .ttl')
+                raise ValueError('file must be .ttl')        
         
-        elif xmpPath is not None:
+        if path is not None:
+            if path.endswith('.JPG') or path.endswith('.PNG'): #data + metadata
+                self.name=ut.get_filename(path)
+                self.timestamp=ut.get_timestamp(path)                
+                self.get_metadata_from_exif_data()
+            else:
+                raise ValueError('file must be .JPG or .PNG')
+       
+        elif img is not None:
+            if type(img) is np.ndarray: #data + metadata
+                self.get_metadata_from_img()
+            else:
+                raise ValueError('variable must be OpenCV image')
+
+        if xmpPath is not None:
             if xmpPath.endswith('.xmp'): #metadata
                 self.name=ut.get_filename(xmpPath)
                 self.timestamp=ut.get_timestamp(xmpPath)
@@ -121,23 +133,11 @@ class ImageNode(Node):
                 self.timestamp=ut.get_timestamp(xmlPath)
                 self.get_metadata_from_xml()
             else:
-                raise ValueError('file must be .xml')
-                
-        elif img is not None:
-            if type(img) is np.ndarray: #data + metadata
-                self.get_metadata_from_img()
-            else:
-                raise ValueError('file must be .ttl')
-        
-        elif path is not None:
-            if path.endswith('.JPG') or path.endswith('.PNG'): #data + metadata
-                self.name=ut.get_filename(path)
-                self.timestamp=ut.get_timestamp(path)                
-                self.get_metadata_from_img()
-            else:
-                raise ValueError('file must be .JPG or .PNG')
+                raise ValueError('file must be .xml')   
+
         if getGeometry:
                 self.get_geometry()
+                self.get_metadata_from_img()
 
     def clear_geometry(self):
         """Clear all geometries in the Node.
@@ -151,6 +151,50 @@ class ImageNode(Node):
         else:
             return None  
 
+    def get_virtual_image(self, geometries: o3d.geometry)-> o3d.cpu.pybind.geometry.Image:
+        pinholeCamera=self.get_pinhole_camera_parameters()
+        if pinholeCamera is not None:
+            return gt.generate_virtual_image(geometries,pinholeCamera)
+        else:
+            return None
+
+    def get_pinhole_camera_parameters(self) -> o3d.camera.PinholeCameraParameters():
+        param=o3d.camera.PinholeCameraParameters()
+        if getattr(self,'cartesianTransform',None) is not None:
+            param.extrinsic=np.linalg.inv(self.cartesianTransform)
+            param.intrinsic=self.get_intrinsic_camera_parameters()
+            self.pinholeCamera=param
+            return self.pinholeCamera
+        else:
+            return None
+
+    def get_intrinsic_camera_parameters(self) -> o3d.camera.PinholeCameraIntrinsic():
+
+        if (getattr(self,'imageWidth',None) is not None and
+            getattr(self,'imageHeight',None) is not None and
+            getattr(self,'focalLength35mm',None) is not None and
+            getattr(self,'resolutionUnit',None) is not None and
+            getattr(self,'xResolution',None) is not None and 
+            getattr(self,'yResolution',None) is not None):
+                width=math.floor(self.imageWidth/(2*self.resolutionUnit))
+                height=math.floor(self.imageHeight/(2*self.resolutionUnit)) 
+                fx=self.focalLength35mm*self.xResolution/self.resolutionUnit
+                fy=self.focalLength35mm*self.yResolution/self.resolutionUnit
+        else:
+            width=640
+            height=480
+            fx=400
+            fy=400
+
+        if (getattr(self,'principalPointU',None) is not None and
+            getattr(self,'principalPointV',None) is not None ):
+            cx=width/2-0.5+self.principalPointU
+            cy=height/2-0.5+self.principalPointV
+        else:
+            cx=width/2-0.5
+            cy=height/2-0.5
+        return o3d.camera.PinholeCameraIntrinsic(width,height,fx,fy,cx,cy)
+
     def get_geometry(self)->bool: 
         """
         get cv2 image from self.path or self.name
@@ -176,60 +220,44 @@ class ImageNode(Node):
                     self.img = cv2.imread(self.path)  
                     return True
         return False
- 
-    def get_exif_data(self) ->bool:
-        """Returns a dictionary from the exif data of an Image item. Also
-        converts the GPS Tags
-        
-        Returns:
-            bool: True if exif data is sucesfully parsed
-        """
-        if getattr(self,'img',None) is not None:
-            exifData = {}
-            info = self.img._getexif()
-            if info:
-                for tag, value in info.items():
-                    decoded = TAGS.get(tag, tag)
-                    if decoded == "GPSInfo":
-                        gps_data = {}
-                        for t in value:
-                            sub_decoded = GPSTAGS.get(t, t)
-                            gps_data[sub_decoded] = value[t]
-
-                        exifData[decoded] = gps_data
-                    else:
-                        exifData[decoded] = value
-                self.exifData=exifData  
-                return True      
-        else:
-            print('No img present')
-            return False        
-
+    
     def get_metadata_from_img(self) -> bool:
+        if getattr(self,'img',None) is not None:
+           self.imageHeight=self.img.shape[0]
+           self.imageWidth=self.img.shape[1]
+           return True
+        else:
+            return False
+
+    def get_metadata_from_exif_data(self) -> bool:
         """Read Metadata from image exif data including GPSInfo if present.
 
         Returns:
             bool: True if metadata is sucesfully parsed
         """
-        self.get_geometry()
-        if getattr(self,'exifData',None) is not None:
-            self.timestamp=ut.get_if_exist(self.exifData, "DateTime")
-            self.xResolution=ut.get_if_exist(self.exifData,"XResolution")
-            self.yResolution=ut.get_if_exist(self.exifData,"YResolution")
-            self.resolutionUnit=ut.get_if_exist(self.exifData,"ResolutionUnit")
-            self.imageWidth=ut.get_if_exist(self.exifData,"ExifImageWidth")
-            self.imageHeight=ut.get_if_exist(self.exifData,"ExifImageHeight")
+ 
+        pix = PIL.Image.open(self.path) 
+        # exifData=pix._getexif()
+        exifData=ut.get_exif_data(pix)#pix._getexif()
+
+        if exifData is not None:
+            self.timestamp=ut.get_if_exist(exifData, "DateTime")
+            self.xResolution=ut.get_if_exist(exifData,"XResolution")
+            self.yResolution=ut.get_if_exist(exifData,"YResolution")
+            self.resolutionUnit=ut.get_if_exist(exifData,"ResolutionUnit")
+            self.imageWidth=ut.get_if_exist(exifData,"ExifImageWidth")
+            self.imageHeight=ut.get_if_exist(exifData,"ExifImageHeight")
             
-            if 'GPSInfo' in self.exifData:
-                gps_info = self.exifData["GPSInfo"]
+            if 'GPSInfo' in exifData:
+                gps_info = exifData["GPSInfo"]
                 if gps_info is not None:
                     # self.GlobalPose=GlobalPose # (structure) SphericalTranslation(lat,long,alt), Quaternion(qw,qx,qy,qz)
                     self.geospatialTransform=[  ut.get_if_exist(gps_info, "GPSLatitude"), 
                                                 ut.get_if_exist(gps_info, "GPSLongitude"),
                                                 ut.get_if_exist(gps_info, "GPSAltitude")]
                     self.coordinateSystem='geospatial-wgs84'
+            return True
         else:
-            print('No exifData in Image')
             return False
     
     def get_metadata_from_xmp(self)->bool:
@@ -241,12 +269,22 @@ class ImageNode(Node):
         if getattr(self,'xmpPath',None) is not None: 
             try:
                 mytree = ET.parse(self.xmpPath)
-                root = mytree.getroot()       
+                root = mytree.getroot()                       
 
                 for img_description in root.iter('{http://www.w3.org/1999/02/22-rdf-syntax-ns#}Description'):
-                    self.focalLength35mm=float(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}FocalLength35mm'])
-                    self.principalPointU=float(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}PrincipalPointU'])
-                    self.principalPointV=float(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}PrincipalPointV'])
+                    lat=ut.xcr_to_lat(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}latitude'])
+                    long=ut.xcr_to_long(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}longitude'])
+                    alt=ut.xcr_to_alt(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}altitude'])
+                    if (lat is not None and
+                        long is not None and
+                        alt is not None):
+                        self.geospatialTransform=np.array([lat, long, alt])
+                    
+                    self.coordinateSystem=img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}Coordinates']
+
+                    self.focalLength35mm=ut.xml_to_float(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}FocalLength35mm'])
+                    self.principalPointU=ut.xml_to_float(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}PrincipalPointU'])
+                    self.principalPointV=ut.xml_to_float(img_description.attrib['{http://www.capturingreality.com/ns/xcr/1.1#}PrincipalPointV'])
 
                     rotationnode=img_description.find('{http://www.capturingreality.com/ns/xcr/1.1#}Rotation')
                     if rotationnode is not None:
@@ -255,16 +293,16 @@ class ImageNode(Node):
 
                     positionnode=img_description.find('{http://www.capturingreality.com/ns/xcr/1.1#}Position')
                     if positionnode is not None:
-                        self.cartesianTransform[0,3]=float(positionnode.text.split(' ')[0])
-                        self.cartesianTransform[1,3]=float(positionnode.text.split(' ')[1])
-                        self.cartesianTransform[2,3]=float(positionnode.text.split(' ')[2])
+                        self.cartesianTransform[0,3]=ut.xml_to_float(positionnode.text.split(' ')[0])
+                        self.cartesianTransform[1,3]=ut.xml_to_float(positionnode.text.split(' ')[1])
+                        self.cartesianTransform[2,3]=ut.xml_to_float(positionnode.text.split(' ')[2])
 
                     coeficientnode=img_description.find('{http://www.capturingreality.com/ns/xcr/1.1#}DistortionCoeficients')
                     if coeficientnode is not None:
                         self.distortionCoeficients=ut.string_to_array(coeficientnode.text)  
                 return True   
             except:
-                print('No exifData in Image')
+                print('XCR parsing error')
                 return False
 
     def get_metadata_from_xml(self) ->bool:
