@@ -234,10 +234,10 @@ def graph_to_nodes(graph : Graph,**kwargs) -> List[Node]:
                     else:
                         setattr(node,attr,(os.getcwd()+'\\'+path)) 
             #INT    
-            elif attr in ['recordCount','faceCount','label','e57Index']:
+            elif attr in ['recordCount','faceCount','label','e57Index','imageWidth','imageHeight','resolutionUnit']:
                 setattr(node,attr,ut.literal_to_int(object)) 
             #FLOAT
-            elif attr in ['xResolution','yResolution','imageWidth','imageHeight','focalLength35mm','principalPointU','principalPointV','accuracy']:
+            elif attr in ['xResolution','yResolution','focalLength35mm','principalPointU','principalPointV','accuracy']:
                 setattr(node,attr,ut.literal_to_float(object)) 
             #ARRAYS
             elif attr in ['hasBIM','hasCAD','hasPcd','hasMesh','hasImg','hasOrtho','hasChild','hasParent','distortionCoeficients']:
