@@ -93,11 +93,17 @@ class MeshNode (GeometryNode):
         if getattr(self,'mesh',None) is not None:
             self.mesh=None
 
-    def get_mesh(self):
-        if getattr(self,'mesh',None) is not None:
+    def get_mesh(self) -> o3d.geometry.TriangleMesh:
+        """Returns the mesh data in the node. If none is present, it will search for the data on drive.
+
+        Returns:
+            o3d.geometry.TriangleMesh or None
+        """
+        if getattr(self,'mesh',None) is not None and len(self.mesh.triangles)>1:
             return self.mesh 
-        else:
-            return None  
+        elif self.get_geometry():
+            return self.mesh
+        return None  
 
     def get_geometry(self)->bool: # deze class is gedeeld met de mesh class
         """
@@ -126,6 +132,9 @@ class MeshNode (GeometryNode):
                         self.path=testPLYPath
                         self.mesh=mesh
                         return True
+        elif getattr(self,'mesh',None) is not None:
+            if len(self.mesh.triangles)>1:
+                return True
         return False
  
     def get_metadata_from_mesh(self) -> bool:
